from datetime import datetime

from .constants import (
    DIFFICULTY_EASY,
    DIFFICULTY_HARD,
    DIFFICULTY_MODERATE,
    DIFFICULTY_VERY_EASY,
    DIFFICULTY_VERY_HARD,
    GUIDE_LABELS,
    UNKNOWN_LOCALE,
)
from .exceptions import UnexpectedDataKindException
from .scraper_generic import ScraperGeneric
from .shared import Global, logger
from .utils import get_api_content, setlocale


class ScraperGuide(ScraperGeneric):
    def __init__(self):
        super().__init__()

    def setup(self):
        self.guide_template = Global.env.get_template("guide.html")

    def get_items_name(self):
        return "guide"

    def _add_guide_to_scrape(self, guideid, locale, is_expected):
        self.add_item_to_scrape(
            guideid,
            {
                "guideid": guideid,
                "locale": locale,
            },
            is_expected,
        )

    def _get_guide_path_from_key(self, guide_key):
        return f"guides/guide_{guide_key}.html"

    def get_guide_link(self, guide):
        guideid = None
        locale = Global.conf.lang_code
        if isinstance(guide, str):
            guideid = guide
        else:
            if "guideid" not in guide or not guide["guideid"]:
                raise UnexpectedDataKindException(
                    f"Impossible to extract guide id from {guide}"
                )
            if "locale" not in guide or not guide["locale"]:
                raise UnexpectedDataKindException(
                    f"Impossible to extract guide locale from {guide}"
                )
            guideid = guide["guideid"]
            locale = guide["locale"]
            # override unknown locale if needed
            if (
                guideid in self.expected_items_keys
                and self.expected_items_keys[guideid]["locale"] == UNKNOWN_LOCALE
            ):
                self.expected_items_keys[guideid]["locale"] = locale
        if not Global.conf.guides and not Global.conf.no_guide:
            self._add_guide_to_scrape(guideid, locale, False)
        return f"../{self._get_guide_path_from_key(guideid)}"

    def build_expected_items(self):
        if Global.conf.no_guide:
            logger.info("No guide required")
            return
        if Global.conf.guides:
            logger.info("Adding required guides as expected")
            for guide in Global.conf.guides:
                self._add_guide_to_scrape(guide, UNKNOWN_LOCALE, True)
            return
        logger.info("Downloading list of guides")
        limit = 200
        offset = 0
        while True:
            guides = get_api_content("/guides", limit=limit, offset=offset)
            if len(guides) == 0:
                break
            for guide in guides:
                guideid = guide["guideid"]
                # Unfortunately for now iFixit API always returns "en" as language on
                # this endpoint, so we consider it as unknown
                self._add_guide_to_scrape(guideid, UNKNOWN_LOCALE, True)
            offset += limit
        logger.info("{} guides found".format(len(self.expected_items_keys)))

    def get_one_item_content(self, item_key, item_data):
        guideid = item_key
        guide = item_data
        locale = guide["locale"]
        if locale == UNKNOWN_LOCALE:
            locale = Global.conf.lang_code  # fallback value
        if locale == "ja":
            locale = "jp"  # Unusual iFixit convention

        guide_content = get_api_content(f"/guides/{guideid}", langid=locale)
        if guide_content is None and locale != "en":
            # guide is most probably available in English anyway
            guide_content = get_api_content(f"/guides/{guideid}", langid="en")

        return guide_content

    def process_one_item(self, item_key, item_data, item_content):
        guide_content = item_content

        if guide_content["type"] != "teardown":
            if guide_content["difficulty"] in DIFFICULTY_VERY_EASY:
                guide_content["difficulty_class"] = "difficulty-1"
            elif guide_content["difficulty"] in DIFFICULTY_EASY:
                guide_content["difficulty_class"] = "difficulty-2"
            elif guide_content["difficulty"] in DIFFICULTY_MODERATE:
                guide_content["difficulty_class"] = "difficulty-3"
            elif guide_content["difficulty"] in DIFFICULTY_HARD:
                guide_content["difficulty_class"] = "difficulty-4"
            elif guide_content["difficulty"] in DIFFICULTY_VERY_HARD:
                guide_content["difficulty_class"] = "difficulty-5"
            else:
                raise UnexpectedDataKindException(
                    "Unknown guide difficulty: '{}' in guide {}".format(
                        guide_content["difficulty"],
                        guide_content["guideid"],
                    )
                )
        with setlocale("en_GB"):
            if guide_content["author"]["join_date"]:
                guide_content["author"]["join_date_rendered"] = datetime.strftime(
                    datetime.fromtimestamp(guide_content["author"]["join_date"]),
                    "%x",
                )
            # TODO: format published date as June 10, 2014 instead of 11/06/2014
            if guide_content["published_date"]:
                guide_content["published_date_rendered"] = datetime.strftime(
                    datetime.fromtimestamp(guide_content["published_date"]),
                    "%x",
                )

        for step in guide_content["steps"]:
            if not step["media"]:
                raise UnexpectedDataKindException(
                    "Missing media attribute in step {} of guide {}".format(
                        step["stepid"], guide_content["guideid"]
                    )
                )
            if step["media"]["type"] not in [
                "image",
                "video",
                "embed",
            ]:
                raise UnexpectedDataKindException(
                    "Unrecognized media type in step {} of guide {}".format(
                        step["stepid"], guide_content["guideid"]
                    )
                )
            if step["media"]["type"] == "video":
                if "data" not in step["media"] or not step["media"]["data"]:
                    raise UnexpectedDataKindException(
                        "Missing 'data' in step {} of guide {}".format(
                            step["stepid"], guide_content["guideid"]
                        )
                    )
                if (
                    "image" not in step["media"]["data"]
                    or not step["media"]["data"]["image"]
                ):
                    raise UnexpectedDataKindException(
                        "Missing outer 'image' in step {} of guide {}".format(
                            step["stepid"], guide_content["guideid"]
                        )
                    )
                if (
                    "image" not in step["media"]["data"]["image"]
                    or not step["media"]["data"]["image"]["image"]
                ):
                    raise UnexpectedDataKindException(
                        "Missing inner 'image' in step {} of guide {}".format(
                            step["stepid"], guide_content["guideid"]
                        )
                    )
            if step["media"]["type"] == "embed":
                if "data" not in step["media"] or not step["media"]["data"]:
                    raise UnexpectedDataKindException(
                        "Missing 'data' in step {} of guide {}".format(
                            step["stepid"], guide_content["guideid"]
                        )
                    )
                if (
                    "html" not in step["media"]["data"]
                    or not step["media"]["data"]["html"]
                ):
                    raise UnexpectedDataKindException(
                        "Missing 'html' in step {} of guide {}".format(
                            step["stepid"], guide_content["guideid"]
                        )
                    )
            for line in step["lines"]:
                if not line["bullet"] in [
                    "black",
                    "red",
                    "orange",
                    "yellow",
                    "green",
                    "blue",
                    "light_blue",
                    "violet",
                    "icon_note",
                    "icon_caution",
                    "icon_caution",
                    "icon_reminder",
                ]:
                    raise UnexpectedDataKindException(
                        "Unrecognized bullet '{}' in step {} of guide {}".format(
                            line["bullet"],
                            step["stepid"],
                            guide_content["guideid"],
                        )
                    )
        guide_rendered = self.guide_template.render(
            guide=guide_content,
            label=GUIDE_LABELS[Global.conf.lang_code],
            metadata=Global.metadata,
        )
        with Global.lock:
            Global.creator.add_item_for(
                path=self._get_guide_path_from_key(item_key),
                title=guide_content["title"],
                content=guide_rendered,
                mimetype="text/html",
                is_front=True,
            )



from .constants import CATEGORY_LABELS
from .exceptions import UnexpectedDataKindException
from .scraper_generic import ScraperGeneric
from .shared import Global, logger
from .utils import get_api_content


class ScraperCategory(ScraperGeneric):
    def __init__(self):
        super().__init__()

    def setup(self):
        self.category_template = Global.env.get_template("category.html")

    def get_items_name(self):
        return "category"

    def _add_category_to_scrape(self, category_key, category_title, is_expected):
        self.add_item_to_scrape(
            category_key,
            {
                "category_title": category_title,
            },
            is_expected,
        )

    def _get_category_key_from_title(self, category_title):
        return Global.convert_title_to_filename(category_title.lower())

    def _get_category_path_from_key(self, category_key):
        return f"categories/category_{category_key}.html"

    def get_category_link(self, category):
        category_title = None
        if isinstance(category, str):
            category_title = category
        elif "title" in category and category["title"]:
            category_title = category["title"]
        else:
            raise UnexpectedDataKindException(
                f"Impossible to extract category title from {category}"
            )
        category_key = self._get_category_key_from_title(category_title)
        if not Global.conf.categories and not Global.conf.no_category:
            self._add_category_to_scrape(category_key, category_title, False)
        return f"../{self._get_category_path_from_key(category_key)}"

    def _process_categories(self, categories):
        for category in categories:
            category_key = self._get_category_key_from_title(category)
            self._add_category_to_scrape(category_key, category, True)
            self._process_categories(categories[category])

    def build_expected_items(self):
        if Global.conf.no_category:
            logger.info("No category required")
            return
        if Global.conf.categories:
            logger.info("Adding required categories as expected")
            for category in Global.conf.categories:
                category_key = self._get_category_key_from_title(category)
                self._add_category_to_scrape(category_key, category, True)
            return
        logger.info("Downloading list of categories")
        categories = get_api_content("/categories", includeStubs=True)
        self._process_categories(categories)
        logger.info("{} categories found".format(len(self.expected_items_keys)))

    def get_one_item_content(self, item_key, item_data):
        categoryid = item_key

        category_content = get_api_content(
            f"/wikis/CATEGORY/{categoryid}", langid=Global.conf.lang_code
        )

        return category_content

    def process_one_item(self, item_key, item_data, item_content):
        category_key = item_key
        category_content = item_content

        category_rendered = self.category_template.render(
            category=category_content,
            label=CATEGORY_LABELS[Global.conf.lang_code],
            metadata=Global.metadata,
            lang=Global.conf.lang_code,
        )

        with Global.lock:
            Global.creator.add_item_for(
                path=self._get_category_path_from_key(category_key),
                title=category_content["display_title"],
                content=category_rendered,
                mimetype="text/html",
                is_front=True,
            )



from .exceptions import UnexpectedDataKindException
from .scraper_generic import ScraperGeneric
from .shared import Global, logger
from .utils import get_api_content


class ScraperInfo(ScraperGeneric):
    def __init__(self):
        super().__init__()

    def setup(self):
        self.info_template = Global.env.get_template("info.html")

    def get_items_name(self):
        return "info"

    def _add_info_to_scrape(self, info_key, info_title, is_expected):
        self.add_item_to_scrape(
            info_key,
            {
                "info_title": info_title,
            },
            is_expected,
        )

    def _get_info_key_from_title(self, info_title):
        return Global.convert_title_to_filename(info_title.lower())

    def _get_info_path_from_key(self, info_key):
        return f"infos/info_{info_key}.html"

    def get_info_link(self, info):
        info_title = None
        if isinstance(info, str):
            info_title = info
        elif "title" in info and info["title"]:
            info_title = info["title"]
        else:
            raise UnexpectedDataKindException(
                f"Impossible to extract info title from {info}"
            )
        info_key = self._get_info_key_from_title(info_title)
        if not Global.conf.infos and not Global.conf.no_info:
            self._add_info_to_scrape(info_key, info_title, False)
        return f"../{self._get_info_path_from_key(info_key)}"

    def build_expected_items(self):
        if Global.conf.no_info:
            logger.info("No info required")
            return
        if Global.conf.infos:
            logger.info("Adding required infos as expected")
            for info_title in Global.conf.infos:
                info_key = self._get_info_key_from_title(info_title)
                self._add_info_to_scrape(info_key, info_title, True)
            return
        logger.info("Downloading list of info")
        limit = 200
        offset = 0
        while True:
            info_wikis = get_api_content("/wikis/INFO", limit=limit, offset=offset)
            if len(info_wikis) == 0:
                break
            for info_wiki in info_wikis:
                info_title = info_wiki["title"]
                info_key = self._get_info_key_from_title(info_title)
                self._add_info_to_scrape(info_key, info_title, True)
            offset += limit
        logger.info("{} info found".format(len(self.expected_items_keys)))

    def get_one_item_content(self, item_key, item_data):
        info_wiki_title = item_key
        info_wiki_content = get_api_content(f"/wikis/INFO/{info_wiki_title}")
        return info_wiki_content

    def process_one_item(self, item_key, item_data, item_content):
        info_wiki_content = item_content

        info_wiki_rendered = self.info_template.render(
            info_wiki=info_wiki_content,
            # label=INFO_WIKI_LABELS[self.conf.lang_code],
            metadata=Global.metadata,
            lang=Global.conf.lang_code,
        )
        with Global.lock:
            Global.creator.add_item_for(
                path=self._get_info_path_from_key(item_key),
                title=info_wiki_content["display_title"],
                content=info_wiki_rendered,
                mimetype="text/html",
                is_front=True,
            )



# -*- coding: utf-8 -*-

import pathlib
import tempfile
import urllib.parse
from dataclasses import dataclass, field
from typing import List, Optional, Set

from zimscraperlib.i18n import get_language_details

ROOT_DIR = pathlib.Path(__file__).parent
NAME = ROOT_DIR.name
DEFAULT_HOMEPAGE = "Main-Page"
UNKNOWN_LOCALE = "unknown"

with open(ROOT_DIR.joinpath("VERSION"), "r") as fh:
    VERSION = fh.read().strip()

SCRAPER = f"{NAME} {VERSION}"

IMAGES_ENCODER_VERSION = 1
URLS = {
    "en": "https://www.ifixit.com",
    "fr": "https://fr.ifixit.com",
    "pt": "https://pt.ifixit.com",
    "de": "https://de.ifixit.com",
    "ru": "https://ru.ifixit.com",
    "ko": "https://ko.ifixit.com",
    "zh": "https://zh.ifixit.com",
    "nl": "https://nl.ifixit.com",
    "ja": "https://jp.ifixit.com",
    "tr": "https://tr.ifixit.com",
    "es": "https://es.ifixit.com",
    "it": "https://it.ifixit.com",
}

DEFAULT_GUIDE_IMAGE_URL = (
    "https://assets.cdn.ifixit.com/static/images/"
    "default_images/GuideNoImage_300x225.jpg"
)

DEFAULT_DEVICE_IMAGE_URL = (
    "https://assets.cdn.ifixit.com/static/images/"
    "default_images/DeviceNoImage_300x225.jpg"
)

DEFAULT_WIKI_IMAGE_URL = (
    "https://assets.cdn.ifixit.com/static/images/"
    "default_images/WikiNoImage_300x225.jpg"
)

# Open this URL in the various languages to retrieve labels below
# https://www.ifixit.com/api/2.0/guides?guideids=219,220,202,206,46465
DIFFICULTY_VERY_EASY = [
    "Very easy",
    "Muito fácil",
    "Très facile",
    "Sehr einfach",
    "Muy fácil",
    "Molto facile",
    "Çok kolay",
    "とても簡単",
    "Zeer eenvoudig",
    "Очень просто",
    "아주 쉬움",
    "非常容易",
]  # guide 219
DIFFICULTY_EASY = [
    "Easy",
    "Fácil",
    "Facile",
    "Einfach",
    "Fácil",
    "Facile",
    "Kolay",
    "簡単",
    "Eenvoudig",
    "Просто",
    "쉬움",
    "简单",
]  # guide 220
DIFFICULTY_MODERATE = [
    "Moderate",
    "Moderado",
    "Modérée",
    "Mittel",
    "Moderado",
    "Moderato",
    "Orta",
    "普通",
    "Gemiddeld",
    "Средняя",
    "보통",
    "中等",
]  # guide 202
DIFFICULTY_HARD = [
    "Difficult",
    "Difícil",
    "Difficile",
    "Schwierig",
    "Difícil",
    "Difficile",
    "Zor",
    "難しい",
    "Moeilijk",
    "Сложно",
    "어려움",
    "困难",
]  # guide 206
DIFFICULTY_VERY_HARD = [
    "Very difficult",
    "Muito difícil",
    "Très difficile",
    "Sehr schwierig",
    "Muy difícil",
    "Molto difficile",
    "Çok zor",
    "とても難しい",
    "Zeer moeilijk",
    "Очень сложно",
    "매우 어려움",
    "非常困难",
]  # guide 46465

# Browse these pages in the various languages to retrieve category + guide labels
# https://www.ifixit.com/Device/Mac
# https://www.ifixit.com/Device/Apple_Watch
# https://www.ifixit.com/Device/Logitech__G502_Hero
# https://www.ifixit.com/Guide/MacBook+Air+11-Inch+Late+2010+Battery+Replacement/4384
# https://www.ifixit.com/Teardown/Apple+Watch+Teardown/40655

HOME_LABELS = {
    "en": {"top_title": "Repair guides for every thing, written by everyone."},
    "fr": {"top_title": "Tutoriels de réparation pour tout, écrits par tous."},
    "pt": {"top_title": "Guias de reparo para tudo, escritos por todos."},
    "de": {"top_title": "Reparaturanleitungen für alles, geschrieben von allen."},
    "ko": {"top_title": "모두가 작성한, 모든 것을 수리하는 안내서."},
    "zh": {"top_title": "大家齐心协力写出的包罗万象的免费修理指南。"},
    "ru": {"top_title": "Руководства по ремонту всего, от всех."},
    "nl": {"top_title": "Reparatiehandleidingen voor alles, door iedereen."},
    "ja": {"top_title": "修理を愛する人たちが作った、あらゆるモノへの修理ガイド"},
    "tr": {"top_title": "Herkes tarafından, her şey için yazılmış tamir kılavuzları."},
    "es": {"top_title": "Guías de reparación para todo, escritas por todos."},
    "it": {"top_title": "Guide di riparazione per ogni cosa, scritte da tutti."},
}
CATEGORY_LABELS = {
    "en": {
        "author": "Author: ",
        "categories_before": "",
        "categories_after": " Categories",
        "featured_guides": "Featured Guides",
        "related_pages": "Related Pages",
        "in_progress_guides": "In Progress Guides",
        "technique_guides": "Techniques",
        "repairability": "Repairability:",
        "replacement_guides": "Replacement Guides",
        "teardown_guides": "Teardowns",
        "disassembly_guides": "Disassembly Guides",
    },
    "fr": {
        "author": "Auteur: ",
        "categories_before": "",
        "categories_after": " catégories",
        "featured_guides": "Tutoriels vedettes",
        "related_pages": "Pages connexes",
        "in_progress_guides": "Tutoriels en cours",
        "technique_guides": "Techniques",
        "repairability": "Réparabilité:",
        "replacement_guides": "Tutoriels de remplacement",
        "teardown_guides": "Vues éclatées",
        "disassembly_guides": "Tutoriels de démontage",
    },
    "pt": {
        "author": "Autor: ",
        "categories_before": "",
        "categories_after": " categorias",
        "featured_guides": "Guia em destaque",
        "technique_guides": "Técnicas",
        "related_pages": "Páginas relacionadas",
        "in_progress_guides": "Guias em andamento",
        "repairability": "Reparabilidade:",
        "replacement_guides": "Guias de reposição",
        "teardown_guides": "Teardowns",
        "disassembly_guides": "Guias de Desmontagem",
    },
    "de": {
        "author": "Autor: ",
        "categories_before": "",
        "categories_after": " Kategorien",
        "featured_guides": "Empfohlene Anleitungen",
        "technique_guides": "Techniken",
        "related_pages": "Verwandte Seiten",
        "in_progress_guides": "Anleitungen in Arbeit",
        "repairability": "Reparierbarkeit:",
        "replacement_guides": "Reparaturanleitungen",
        "teardown_guides": "Teardowns",
        "disassembly_guides": "Demontageanleitungen",
    },
    "ko": {
        "author": "작성자: ",
        "categories_before": "범주 ",
        "categories_after": "개",
        "featured_guides": "주요 가이드",
        "technique_guides": "기술",
        "related_pages": "관련 페이지",
        "in_progress_guides": "작성 중인 안내서",
        "repairability": "수리 용이성:",
        "replacement_guides": "교체 가이드",
        "teardown_guides": "분해도",
        "disassembly_guides": "분해 안내서",
    },
    "zh": {
        "author": "作者: ",
        "categories_before": "",
        "categories_after": " 个类别",
        "featured_guides": "精选指南",
        "technique_guides": "技术",
        "related_pages": "相关页面",
        "in_progress_guides": "正在编写中的指南",
        "repairability": "可修复性:",
        "replacement_guides": "更换指南",
        "teardown_guides": "拆​解",
        "disassembly_guides": "拆卸指南",
    },
    "ru": {
        "author": "Автор: ",
        "categories_before": "",
        "categories_after": " категорий",
        "featured_guides": "Связанные инструкции",
        "technique_guides": "Техника",
        "related_pages": "Связанные страницы",
        "in_progress_guides": "Многоступенчатые руководства",
        "repairability": "Ремонтопригодность:",
        "replacement_guides": "Инструкции по заменам",
        "teardown_guides": "Разбираем",
        "disassembly_guides": "Инструкции по демонтажу",
    },
    "nl": {
        "author": "Auteur: ",
        "categories_before": "",
        "categories_after": " Categorieën",
        "featured_guides": "Aanbevolen Handleidingen",
        "technique_guides": "Technieken",
        "related_pages": "Gerelateerde pagina's",
        "in_progress_guides": "Handleidingen in wording",
        "repairability": "Repareerbaarheid:",
        "replacement_guides": "Vervangingshandleidingen",
        "teardown_guides": "Demontages",
        "disassembly_guides": "Demontagehandleidingen",
    },
    "ja": {
        "author": "作成者: ",
        "categories_before": "",
        "categories_after": " カテゴリー",
        "featured_guides": "おすすめのガイド",
        "technique_guides": "テクニック",
        "related_pages": "関連ページ",
        "in_progress_guides": "作成中のガイド",
        "repairability": "リペアビリティ:",
        "replacement_guides": "交換ガイド",
        "teardown_guides": "分解",
        "disassembly_guides": "分解ガイド",
    },
    "tr": {
        "author": "Yazar: ",
        "categories_before": "",
        "categories_after": " Kategori",
        "featured_guides": "Featured Guides",  # not present for now on website ...
        "technique_guides": "Teknikler",
        "related_pages": "İlgili Sayfalar",
        "in_progress_guides": "Yapım Aşamasındaki Kılavuzlar",
        "repairability": "Onarılabilirlik:",
        "replacement_guides": "Parça Değişim Kılavuzları",
        "teardown_guides": "Teardown'lar",
        "disassembly_guides": "Söküm Kılavuzları",
    },
    "it": {
        "author": "Autore: ",
        "categories_before": "",
        "categories_after": " Categorie",
        "featured_guides": "Guide In Evidenza",
        "technique_guides": "Tecniche",
        "related_pages": "Pagine Correlate",
        "in_progress_guides": "Guide in lavorazione",
        "repairability": "Riparabilità:",
        "replacement_guides": "Guide Sostituzione",
        "teardown_guides": "Smontaggi",
        "disassembly_guides": "Guide di smontaggio",
    },
    "es": {
        "author": "Autor/a: ",
        "categories_before": "",
        "categories_after": " Categorías",
        "featured_guides": "Guías Destacadas",
        "technique_guides": "Técnicas",
        "related_pages": "Páginas Relacionadas",
        "in_progress_guides": "Guías en Progreso",
        "repairability": "Reparabilidad:",
        "replacement_guides": "Guías de reemplazo",
        "teardown_guides": "Desmontajes",
        "disassembly_guides": "Guías de Desmontaje",
    },
}

GUIDE_LABELS = {
    "en": {
        "written_by": "Written By:",
        "difficulty": "Difficulty",
        "steps": "Steps",
        "time_required": " Time Required",
        "sections": "Sections",
        "flags": "Flags",
        "introduction": "Introduction",
        "step_no_before": "Step ",
        "step_no_after": "",
        "conclusion": "Conclusion",
        "author": "Author",
        "reputation": "Reputation",
        "member_since_before": "Member since: ",
        "member_since_after": "",
        "published": "Published: ",
        "teardown": "Teardown",
    },
    "fr": {
        "written_by": "Rédigé par :",
        "difficulty": "Difficulté",
        "steps": "Étapes",
        "time_required": "Temps nécessaire",
        "sections": "Sections",
        "flags": "Drapeaux",
        "introduction": "Introduction",
        "step_no_before": "Étape ",
        "step_no_after": "",
        "conclusion": "Conclusion",
        "author": "Auteur",
        "reputation": "Réputation",
        "member_since_before": "Membre depuis le ",
        "member_since_after": "",
        "published": "Publication : ",
        "teardown": "Vue éclatée",
    },
    "pt": {
        "written_by": "Escrito Por:",
        "difficulty": "Dificuldade",
        "steps": "Passos",
        "time_required": "Tempo necessário",
        "sections": "Partes",
        "flags": "Sinalizadores",
        "introduction": "Introdução",
        "step_no_before": "Passo ",
        "step_no_after": "",
        "conclusion": "Conclusão",
        "author": "Autor(a)",
        "reputation": "Reputação",
        "member_since_before": "Membro desde: ",
        "member_since_after": "",
        "published": "Publicado em: ",
        "teardown": "Teardown",
    },
    "de": {
        "written_by": "Geschrieben von:",
        "difficulty": "Schwierigkeitsgrad",
        "steps": "Schritte",
        "time_required": "Zeitaufwand",
        "sections": "Abschnitte",
        "flags": "Kennzeichnungen",
        "introduction": "Einleitung",
        "step_no_before": "Schritt ",
        "step_no_after": "",
        "conclusion": "Abschluss",
        "author": "Tinte & Feder",
        "reputation": "Reputation",
        "member_since_before": "Mitglied seit: ",
        "member_since_after": "",
        "published": "Veröffentlicht: ",
        "teardown": "Teardown",
    },
    "ko": {
        "written_by": "작성자",
        "difficulty": "난이도",
        "steps": " 단계",
        "time_required": " 소요 시간",
        "sections": "섹션",
        "flags": "플래그",
        "introduction": "소개",
        "step_no_before": "",
        "step_no_after": " 단계",
        "conclusion": "결론",
        "author": "작성자",
        "reputation": "평판",
        "member_since_before": "회원 가입일: ",
        "member_since_after": "",
        "published": "게시일: ",
        "teardown": "분해도",
    },
    "zh": {
        "written_by": "撰写者：",
        "difficulty": "难度",
        "steps": "步骤",
        "time_required": "所需时间",
        "sections": " 节",
        "flags": " 标志",
        "introduction": "简介",
        "step_no_before": "步骤 1",
        "step_no_after": "",
        "conclusion": " 结论",
        "author": "作者",
        "reputation": "信誉积分",
        "member_since_before": "",
        "member_since_after": "日注册",
        "published": "发布于：",
        "teardown": " 拆解",
    },
    "ru": {
        "written_by": "Автор:",
        "difficulty": "Сложность",
        "steps": "Шаги",
        "time_required": "Необходимое время",
        "sections": "Разделы",
        "flags": "Флаги",
        "introduction": "Введение",
        "step_no_before": "Шаг ",
        "step_no_after": "",
        "conclusion": "Заключение",
        "author": "Автор",
        "reputation": "Репутация",
        "member_since_before": "Участник с: ",
        "member_since_after": "",
        "published": "Опубликовано: ",
        "teardown": "Разбираем",
    },
    "nl": {
        "written_by": "Geschreven door:",
        "difficulty": "Moeilijkheid",
        "steps": "Stappen",
        "time_required": " Tijd vereist",
        "sections": "Secties",
        "flags": "Markeringen",
        "introduction": "Inleiding",
        "step_no_before": "Stap ",
        "step_no_after": "",
        "conclusion": "Conclusie",
        "author": "Auteur",
        "reputation": "Reputatie",
        "member_since_before": "Lid sinds: ",
        "member_since_after": "",
        "published": "Gepubliceerd: ",
        "teardown": "Uit elkaar gehaald",
    },
    "ja": {
        "written_by": "作成者：",
        "difficulty": "難易度",
        "steps": "手順",
        "time_required": "所要時間",
        "sections": "セクション",
        "flags": "フラグ",
        "introduction": "はじめに",
        "step_no_before": "手順 ",
        "step_no_after": "",
        "conclusion": "終わりに",
        "author": "作成者",
        "reputation": "ポイント",
        "member_since_before": "メンバー登録日: ",
        "member_since_after": "",
        "published": "公開日: ",
        "teardown": "分解",
    },
    "tr": {
        "written_by": "Yazan:",
        "difficulty": "Zorluk",
        "steps": " Adımlar",
        "time_required": "Gerekli Süre",
        "sections": "Bölümler",
        "flags": "İşaretler",
        "introduction": "Giriş",
        "step_no_before": "Adım ",
        "step_no_after": "",
        "conclusion": "Sonuç",
        "author": "Yazar",
        "reputation": "İtibar",
        "member_since_before": "Üyelik tarihi: ",
        "member_since_after": "",
        "published": "Yayımlama: ",
        "teardown": "Teardown",
    },
    "it": {
        "written_by": "Scritto Da:",
        "difficulty": "Difficoltà",
        "steps": "Passi",
        "time_required": "Tempo Necessario",
        "sections": "Sezioni",
        "flags": "Contrassegni",
        "introduction": "Introduzione",
        "step_no_before": "Passo ",
        "step_no_after": "",
        "conclusion": "Conclusione",
        "author": "Autore",
        "reputation": "Reputazione",
        "member_since_before": "Membro da: ",
        "member_since_after": "",
        "published": "Pubblicato: ",
        "teardown": "Smontaggio",
    },
    "es": {
        "written_by": "Escrito Por:",
        "difficulty": "Dificultad",
        "steps": "Pasos",
        "time_required": "Tiempo Requerido",
        "sections": "Secciones",
        "flags": "Banderas",
        "introduction": "Introducción",
        "step_no_before": "Paso ",
        "step_no_after": "",
        "conclusion": "Conclusión",
        "author": "Autor",
        "reputation": "Reputación",
        "member_since_before": "Miembro Desde ",
        "member_since_after": "",
        "published": "Publicado: ",
        "teardown": "Desmontaje",
    },
}


API_PREFIX = "/api/2.0"


@dataclass
class Conf:
    required = [
        "lang_code",
        "output_dir",
    ]

    lang_code: str = ""
    language: dict = field(default_factory=dict)
    main_url: str = ""

    # zim params
    name: str = ""
    title: Optional[str] = ""
    description: Optional[str] = ""
    author: Optional[str] = ""
    publisher: Optional[str] = ""
    fname: Optional[str] = ""
    tag: List[str] = field(default_factory=list)

    # customization
    icon: Optional[str] = ""
    categories: Set[str] = field(default_factory=set)
    no_category: Optional[bool] = False
    guides: Set[str] = field(default_factory=set)
    no_guide: Optional[bool] = False
    infos: Set[str] = field(default_factory=set)
    no_info: Optional[bool] = False

    # filesystem
    _output_dir: Optional[str] = "."
    _tmp_dir: Optional[str] = "."
    output_dir: Optional[pathlib.Path] = None
    tmp_dir: Optional[pathlib.Path] = None

    # performances
    nb_threads: Optional[int] = -1
    s3_url_with_credentials: Optional[str] = ""

    # error handling
    max_missing_items_percent: Optional[int] = 0
    max_error_items_percent: Optional[int] = 0

    # debug/devel
    build_dir_is_tmp_dir: Optional[bool] = False
    keep_build_dir: Optional[bool] = False
    scrape_only_first_items: Optional[bool] = False
    debug: Optional[bool] = False
    delay: Optional[float] = 0
    api_delay: Optional[float] = 0
    cdn_delay: Optional[float] = 0
    stats_filename: Optional[str] = None
    skip_checks: Optional[bool] = False

    @staticmethod
    def get_url(lang_code: str) -> urllib.parse.ParseResult:
        return urllib.parse.urlparse(URLS[lang_code])

    @property
    def domain(self) -> str:
        return self.main_url.netloc

    @property
    def api_url(self) -> str:
        return self.main_url + API_PREFIX

    @property
    def s3_url(self) -> str:
        return self.s3_url_with_credentials

    def __post_init__(self):
        self.main_url = Conf.get_url(self.lang_code)
        self.language = get_language_details(self.lang_code)
        self.output_dir = pathlib.Path(self._output_dir).expanduser().resolve()
        self.output_dir.mkdir(parents=True, exist_ok=True)

        self.tmp_dir = pathlib.Path(self._tmp_dir).expanduser().resolve()
        self.tmp_dir.mkdir(parents=True, exist_ok=True)
        if self.build_dir_is_tmp_dir:
            self.build_dir = self.tmp_dir
        else:
            self.build_dir = pathlib.Path(
                tempfile.mkdtemp(prefix=f"ifixit_{self.lang_code}_", dir=self.tmp_dir)
            )

        if self.stats_filename:
            self.stats_filename = pathlib.Path(self.stats_filename).expanduser()
            self.stats_filename.parent.mkdir(parents=True, exist_ok=True)

        # support semi-colon separated tags as well
        if self.tag:
            for tag in self.tag.copy():
                if ";" in tag:
                    self.tag += [p.strip() for p in tag.split(";")]
                    self.tag.remove(tag)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4 nu
# pylint: disable=cyclic-import

import datetime
import logging
import re
import threading

from jinja2 import Environment, FileSystemLoader, select_autoescape
from zimscraperlib.logging import getLogger as lib_getLogger
from zimscraperlib.zim.creator import Creator

from .constants import (
    DEFAULT_DEVICE_IMAGE_URL,
    DEFAULT_GUIDE_IMAGE_URL,
    DEFAULT_HOMEPAGE,
    DEFAULT_WIKI_IMAGE_URL,
    NAME,
    ROOT_DIR,
)


class ImageUrlNotFound(Exception):
    pass


class Global:
    """Shared context accross all scraper components"""

    debug = False
    logger = lib_getLogger(
        NAME,
        level=logging.INFO,
        log_format="[%(threadName)s::%(asctime)s] %(levelname)s:%(message)s",
    )
    conf = None

    metadata = {}

    creator = None
    imager = None
    env = None
    lock = threading.Lock()

    @staticmethod
    def set_debug(value):
        Global.debug = value
        level = logging.DEBUG if value else logging.INFO
        Global.logger.setLevel(level)
        for handler in Global.logger.handlers:
            handler.setLevel(level)

    @staticmethod
    def setup():
        # order matters are there are references between them

        # images handled on a different queue.
        # mostly network I/O to retrieve and/or upload image.
        # if not in S3 bucket, convert/optimize webp image
        # svg images, stored but not optimized
        from .executor import Executor

        Global.img_executor = Executor(
            queue_size=20,
            nb_workers=10,
            prefix="IMG-T-",
        )

        from .imager import Imager

        Global.imager = Imager()

        Global.creator = Creator(
            filename=Global.conf.output_dir.joinpath(Global.conf.fname),
            main_path=DEFAULT_HOMEPAGE,
            favicon_path="illustration",
            language=Global.conf.language["iso-639-3"],
            title=Global.conf.title,
            description=Global.conf.description,
            creator=Global.conf.author,
            publisher=Global.conf.publisher,
            name=Global.conf.name,
            tags=";".join(Global.conf.tags),
            date=datetime.date.today(),
        ).config_verbose(True)

        # jinja2 environment setup
        Global.env = Environment(
            loader=FileSystemLoader(ROOT_DIR.joinpath("templates")),
            autoescape=select_autoescape(),
        )
        Global.env.filters["guides_in_progress"] = Global.guides_in_progress
        Global.env.filters["get_image_path"] = Global.get_image_path
        Global.env.filters["get_image_url"] = Global.get_image_url
        Global.env.filters["cleanup_rendered_content"] = Global.cleanup_rendered_content

    @staticmethod
    def guides_in_progress(guides, in_progress=True):
        if in_progress:
            return [guide for guide in guides if "GUIDE_IN_PROGRESS" in guide["flags"]]
        return [guide for guide in guides if "GUIDE_IN_PROGRESS" not in guide["flags"]]

    @staticmethod
    def get_image_path(image_url):
        return f"../{Global.imager.defer(url=image_url)}"

    @staticmethod
    def _get_image_url_search(obj, for_guide, for_device, for_wiki):
        if "standard" in obj:
            return obj["standard"]
        if "medium" in obj:
            return obj["medium"]
        if "large" in obj:
            return obj["large"]
        if "original" in obj:
            return obj["original"]
        if for_guide:
            return DEFAULT_GUIDE_IMAGE_URL
        if for_device:
            return DEFAULT_DEVICE_IMAGE_URL
        if for_wiki:
            return DEFAULT_WIKI_IMAGE_URL
        raise ImageUrlNotFound(f"Unable to find image URL in object {obj}")

    @staticmethod
    def get_image_url(obj, for_guide=False, for_device=False, for_wiki=False):
        if "image" in obj and obj["image"]:
            return Global._get_image_url_search(
                obj["image"], for_guide, for_device, for_wiki
            )
        return Global._get_image_url_search(obj, for_guide, for_device, for_wiki)

    guide_regex_full = re.compile(
        r"href=\"https://\w*\.ifixit\.\w*/Guide/.*/(?P<guide_id>\d*)\""
    )
    guide_regex_rel = re.compile(r"href=\"/Guide/.*/(?P<guide_id>\d*).*?\"")

    gbl_image_regex = r"<img(?P<image_before>.*?)src\s*=\s*\"(?P<image_url>.*?)\""
    gbl_href_regex = r"href\s*=\s*\"(?P<href_url>.*?)\""
    gbl_regex = re.compile(f"{gbl_image_regex}|{gbl_href_regex}")

    href_anchor_regex = r"^(?P<anchor>#.*)$"
    href_object_kind_regex = (
        r"^(?:https*://[\w\.]*(?:ifixit)[\w\.]*)*/"
        r"(?P<kind>Device|Topic|User|Team|Info|Wiki|Store|Boutique|Tienda|Guide|"
        r"Anleitung|Guía|Guida|Tutoriel).*/(?P<object>[\w%_-]*)(?P<after>.*)$"
    )
    href_regex = re.compile(
        f"{href_anchor_regex}|{href_object_kind_regex}", flags=re.IGNORECASE
    )

    @staticmethod
    def _process_href_regex(href):
        if "Guide/login/register" in href:
            return "../home/placeholder.html"
        if "Guide/new" in href:
            return "../home/placeholder.html"

        found_none = True
        found_one = False
        for match in Global.href_regex.finditer(href):
            if found_one:
                logger.warn(f"Too many matches in _process_href_regex for '{href}'")
                return href
            found_one = True
            found_none = False
            if match.group("anchor"):
                return f"ANCHOR_{match.group('anchor')}"
            if match.group("kind"):
                if match.group("kind").lower() in ["device", "topic"]:
                    return (
                        f"{Global.get_category_link(match.group('object'))}"
                        f"{match.group('after')}"
                    )
                if match.group("kind").lower() in ["info"]:
                    return (
                        f"{Global.get_info_link(match.group('object'))}"
                        f"{match.group('after')}"
                    )
                if match.group("kind").lower() in ["user", "team", "info", "wiki"]:
                    return "../home/placeholder.html"
                if match.group("kind").lower() in ["store", "boutique", "tienda"]:
                    return "../home/placeholder.html"
                if match.group("kind").lower() in [
                    "guide",
                    "anleitung",
                    "guía",
                    "guida",
                    "tutoriel",
                ]:
                    return (
                        f"{Global.get_guide_link(match.group('object'))}"
                        f"{match.group('after')}"
                    )
                raise Exception(
                    f"Unsupported kind '{match.group('kind')}'"
                    " in _process_href_regex"
                )
            raise Exception("Unsupported match in _process_href_regex")
        if found_none:
            return href

    @staticmethod
    def _process_gbl_regex(match):
        if match.group("image_url"):
            return (
                f"<img{match.group('image_before')}"
                f"src=\"{Global.get_image_path(match.group('image_url'))}\""
            )
        if match.group("href_url"):
            return f"href=\"{Global._process_href_regex(match.group('href_url'))}\""
        raise Exception("Unsupported match in cleanup_rendered_content")

    @staticmethod
    def cleanup_rendered_content(content):
        return re.sub(Global.gbl_regex, Global._process_gbl_regex, content)

    @staticmethod
    def convert_title_to_filename(title):
        return re.sub(r"\s", "_", title)


class GlobalMixin:
    @property
    def conf(self):
        return Global.conf

    @property
    def metadata(self):
        return Global.metadata

    @property
    def creator(self):
        return Global.creator

    @property
    def lock(self):
        return Global.lock

    @property
    def imager(self):
        return Global.imager

    @property
    def executor(self):
        return Global.executor

    @property
    def env(self):
        return Global.env

    @property
    def info_wiki_template(self):
        return Global.info_wiki_template


logger = Global.logger



# -*- coding: utf-8 -*-

import json
import re

from .constants import DEFAULT_HOMEPAGE, HOME_LABELS
from .exceptions import CategoryHomePageContentError
from .scraper_generic import ScraperGeneric
from .shared import Global, logger
from .utils import get_soup


class ScraperHomepage(ScraperGeneric):
    def __init__(self):
        super().__init__()

    def setup(self):
        self.homepage_template = Global.env.get_template("home.html")
        self.placeholder_template = Global.env.get_template("placeholder.html")

    def get_items_name(self):
        return "home"

    def build_expected_items(self):
        self.add_item_to_scrape(1, 1, True)

    def get_one_item_content(self, item_key, item_data):
        soup, _ = get_soup("/Guide")
        return soup

    def process_one_item(self, item_key, item_data, item_content):
        soup = item_content

        # extract and clean main content
        home_content = {
            "main_title": soup.find("title").string,
            "page_title": self._extract_page_title_from_page(soup),
            "primary_title": self._extract_primary_title_from_page(soup),
            "secondary_title": self._extract_secondary_title_from_page(soup),
            "callout": self._extract_callout_from_page(soup),
            "featured_categories": self._extract_featured_categories_from_page(soup),
            "sub_categories": self._extract_sub_categories_from_page(soup),
        }

        for fc in home_content["featured_categories"]:
            image_path = Global.imager.defer(url=fc["img_url"])
            fc["img_path"] = image_path

        home_content["callout"]["img_path"] = Global.imager.defer(
            url=home_content["callout"]["img_url"]
        )

        logger.debug(
            "Content extracted from /Guide:\n" f"{json.dumps(home_content,indent=2)}"
        )

        homepage = self.homepage_template.render(
            home_content=home_content,
            metadata=Global.metadata,
            label=HOME_LABELS[Global.conf.lang_code],
        )

        placeholder = self.placeholder_template.render(
            home_content=home_content,
            metadata=Global.metadata,
            label=HOME_LABELS[Global.conf.lang_code],
        )

        with Global.lock:
            Global.creator.add_item_for(
                path="home/home.html",
                title=Global.conf.title,
                content=homepage,
                mimetype="text/html",
                is_front=True,
            )

            Global.creator.add_redirect(
                path=DEFAULT_HOMEPAGE, target_path="home/home.html"
            )

            Global.creator.add_item_for(
                path="home/placeholder.html",
                title="Placeholder",
                content=placeholder,
                mimetype="text/html",
                is_front=False,
            )

    _device_link_regex_without_href = re.compile(r"/Device/(?P<device>.*)")

    def _extract_page_title_from_page(self, soup):
        page_title_selector = "h1.page-title span"
        p = soup.select(page_title_selector)
        if len(p) == 0:
            raise CategoryHomePageContentError(
                "No text found in page with selector " f"'{page_title_selector}'"
            )
        if len(p) > 1:
            raise CategoryHomePageContentError(
                "Too many text found in page with selector " f"'{page_title_selector}'"
            )
        text = p[0].text
        if len(text) == 0:
            raise CategoryHomePageContentError(
                "Empty text found in page with selector " f"'{page_title_selector}'"
            )
        return text

    def _extract_primary_title_from_page(self, soup):
        primary_title_selector = "div.primary-divider p"
        p = soup.select(primary_title_selector)
        if len(p) == 0:
            raise CategoryHomePageContentError(
                "No text found in page with selector " f"'{primary_title_selector}'"
            )
        if len(p) > 1:
            raise CategoryHomePageContentError(
                "Too many text found in page with selector "
                f"'{primary_title_selector}'"
            )
        text = p[0].text
        if len(text) == 0:
            raise CategoryHomePageContentError(
                "Empty text found in page with selector " f"'{primary_title_selector}'"
            )
        return text

    def _extract_secondary_title_from_page(self, soup):
        secondary_title_selector = "div.secondary-divider p"
        p = soup.select(secondary_title_selector)
        if len(p) == 0:
            raise CategoryHomePageContentError(
                "No text found in page with selector " f"'{secondary_title_selector}'"
            )
        if len(p) > 1:
            raise CategoryHomePageContentError(
                "Too many text found in page with selector "
                f"'{secondary_title_selector}'"
            )
        text = p[0].text
        if len(text) == 0:
            raise CategoryHomePageContentError(
                "Empty text found in page with selector "
                f"'{secondary_title_selector}'"
            )
        return text

    def _extract_callout_from_page(self, soup):
        return {
            "content": self._extract_callout_content_from_page(soup),
            "img_url": self._extract_callout_img_src_from_page(soup),
        }

    def _extract_callout_content_from_page(self, soup):
        page_callout_selector = "div.page-callout-content"
        p = soup.select(page_callout_selector)
        if len(p) == 0:
            raise CategoryHomePageContentError(
                "No callout content found in page with selector "
                f"'{page_callout_selector}'"
            )
        if len(p) > 1:
            raise CategoryHomePageContentError(
                "Too many callout content found in page with selector "
                f"'{page_callout_selector}'"
            )
        return f"{p[0]}"

    def _extract_callout_img_src_from_page(self, soup):
        page_callout_selector = "div.page-callout-inner img"
        p = soup.select(page_callout_selector)
        if len(p) == 0:
            raise CategoryHomePageContentError(
                "No callout img found in page with selector "
                f"'{page_callout_selector}'"
            )
        if len(p) > 1:
            raise CategoryHomePageContentError(
                "Too many callout img found in page with selector "
                f"'{page_callout_selector}'"
            )
        src = p[0].attrs.get("src")
        if len(src) == 0:
            raise CategoryHomePageContentError(
                "Empty img src found in featured callout with selector "
                f"'{page_callout_selector}'"
            )
        return src

    def _extract_featured_categories_from_page(self, soup):
        featured_categories_css_selector = "a.featured-category-item"
        featured_categories = [
            {
                "text": self._extract_text_from_featured_category(fc),
                "img_url": self._extract_img_src_from_featured_category(fc),
                "name": self._extract_name_from_featured_category(fc),
                "title": self._extract_title_from_featured_category(fc),
            }
            for fc in soup.select(featured_categories_css_selector)
        ]

        if len(featured_categories) == 0:
            raise CategoryHomePageContentError(
                "No featured categories found with selector "
                f"'{featured_categories_css_selector}'"
            )
        return featured_categories

    def _extract_text_from_featured_category(self, fc):
        featured_category_text_css_selector = "p.featured-category-title"
        p = fc.select(featured_category_text_css_selector)
        if len(p) == 0:
            raise CategoryHomePageContentError(
                "No text found in featured category with selector "
                f"'{featured_category_text_css_selector}'"
            )
        if len(p) > 1:
            raise CategoryHomePageContentError(
                "Too many text found in featured category with selector "
                f"'{featured_category_text_css_selector}'"
            )
        text = p[0].text
        if len(text) == 0:
            raise CategoryHomePageContentError(
                "Empty text found in featured category with selector "
                f"'{featured_category_text_css_selector}'"
            )
        return text

    def _extract_img_src_from_featured_category(self, fc):
        featured_category_img_css_selector = "img"
        p = fc.select(featured_category_img_css_selector)
        if len(p) == 0:
            raise CategoryHomePageContentError(
                "No img found in featured category with selector "
                f"'{featured_category_img_css_selector}'"
            )
        if len(p) > 1:
            raise CategoryHomePageContentError(
                "Too many img found in featured category with selector "
                f"'{featured_category_img_css_selector}'"
            )
        src = p[0].attrs.get("src")
        if len(src) == 0:
            raise CategoryHomePageContentError(
                "Empty img src found in featured category with selector "
                f"'{featured_category_img_css_selector}'"
            )
        return src

    def _extract_name_from_featured_category(self, fc):
        href = fc.attrs.get("href")
        if len(href) == 0:
            raise CategoryHomePageContentError("Empty href found in featured category")
        name = self._device_link_regex_without_href.sub("\\g<device>", href)
        if name == href:
            raise CategoryHomePageContentError(
                f"Extracting name from featured category failed ; href:'{href}'"
            )
        return name

    def _extract_title_from_featured_category(self, fc):
        title = fc.attrs.get("title")
        if len(title) == 0:
            raise CategoryHomePageContentError("Empty title found in featured category")
        return title

    def _extract_sub_categories_from_page(self, soup):
        sub_categories_css_selector = "a.sub-category"
        sub_categories = [
            {
                "text": self._extract_text_from_sub_category(fc),
                "name": self._extract_name_from_sub_category(fc),
                "count": self._extract_count_from_sub_category(fc),
                "title": self._extract_title_from_sub_category(fc),
            }
            for fc in soup.select(sub_categories_css_selector)
        ]
        if len(sub_categories) == 0:
            raise CategoryHomePageContentError(
                "No sub-categories found with selector "
                f"'{sub_categories_css_selector}'"
            )
        return sub_categories

    def _extract_text_from_sub_category(self, sc):
        sub_category_text_css_selector = "span.sub-category-title-text"
        p = sc.select(sub_category_text_css_selector)
        if len(p) == 0:
            raise CategoryHomePageContentError(
                "No text found in sub-category with selector "
                f"'{sub_category_text_css_selector}'"
            )
        if len(p) > 1:
            raise CategoryHomePageContentError(
                "Too many text found in sub-category with selector "
                f"'{sub_category_text_css_selector}'"
            )
        text = p[0].text
        if len(text) == 0:
            raise CategoryHomePageContentError(
                "Empty text found in sub-category with selector "
                f"'{sub_category_text_css_selector}'"
            )
        return text

    def _extract_name_from_sub_category(self, sc):
        href = sc.attrs.get("href")
        if len(href) == 0:
            raise CategoryHomePageContentError("Empty href found in sub-category")
        name = self._device_link_regex_without_href.sub("\\g<device>", href)
        if name == href:
            raise CategoryHomePageContentError(
                f"Extracting name from sub-category failed ; href:'{href}'"
            )
        return name

    def _extract_count_from_sub_category(self, sc):
        sub_category_img_css_selector = "span.overflow-slide-in"
        p = sc.select(sub_category_img_css_selector)
        if len(p) == 0:
            raise CategoryHomePageContentError(
                "No span found in sub-category with selector "
                f"'{sub_category_img_css_selector}'"
            )
        if len(p) > 1:
            raise CategoryHomePageContentError(
                "Too many span found in sub-category with selector "
                f"'{sub_category_img_css_selector}'"
            )
        text = p[0].text
        if len(text) == 0:
            raise CategoryHomePageContentError(
                "Empty span text found in sub-category with selector "
                f"'{sub_category_img_css_selector}'"
            )
        try:
            return int(text)
        except ValueError:
            raise CategoryHomePageContentError(
                f"Failed to convert span text '{text}' to integer for " "sub-category"
            )

    def _extract_title_from_sub_category(self, sc):
        sub_category_img_css_selector = "span.overflow-slide-in"
        p = sc.select(sub_category_img_css_selector)
        if len(p) == 0:
            raise CategoryHomePageContentError(
                "No span found in sub-category with selector "
                f"'{sub_category_img_css_selector}'"
            )
        if len(p) > 1:
            raise CategoryHomePageContentError(
                "Too many span found in sub-category with selector "
                f"'{sub_category_img_css_selector}'"
            )
        title = p[0].attrs.get("title")
        if len(title) == 0:
            raise CategoryHomePageContentError(
                "Empty span title found in sub-category with selector "
                f"'{sub_category_img_css_selector}'"
            )
        return title

    def _extract_footer_stats_from_page(self, soup):
        footer_stats_css_selector = "div.footer-stats div"
        footer_stats = [
            self._extract_details_from_footer_stats(fc)
            for fc in soup.select(footer_stats_css_selector)
        ]
        if len(footer_stats) == 0:
            raise CategoryHomePageContentError(
                "No footer stats found with selector " f"'{footer_stats_css_selector}'"
            )
        return footer_stats

    def _extract_details_from_footer_stats(self, fs):
        footer_stats_text_css_selector = "p"
        p = fs.select(footer_stats_text_css_selector)
        if len(p) == 0:
            raise CategoryHomePageContentError(
                "No text found in footer stat with selector "
                f"'{footer_stats_text_css_selector}'"
            )
        if len(p) == 1:
            raise CategoryHomePageContentError(
                "Insufficient text found in footer stat with selector "
                f"'{footer_stats_text_css_selector}'"
            )
        if len(p) > 2:
            raise CategoryHomePageContentError(
                "Too many text found in footer stat with selector "
                f"'{footer_stats_text_css_selector}'"
            )
        text0 = p[0].text
        if len(text0) == 0:
            raise CategoryHomePageContentError(
                "Empty text found first paragraph of footer stat "
                f"'{footer_stats_text_css_selector}'"
            )
        text0digits = re.sub("[^0-9]", "", text0)
        if len(text0digits) == 0:
            raise CategoryHomePageContentError(
                f"No digits found in text '{text0}' of footer stat"
            )
        text1 = p[1].text
        if len(text1) == 0:
            raise CategoryHomePageContentError(
                "Empty text found second paragraph of footer stat "
                f"'{footer_stats_text_css_selector}'"
            )
        try:
            return {
                "value": int(text0digits),
                "text": text1,
            }
        except ValueError:
            raise CategoryHomePageContentError(
                f"Failed to convert text '{text0digits}' to integer for footer " "stat"
            )

    def _extract_footer_copyright_from_page(self, soup):
        footer_copyright_css_selector = "div.footer-container div.copyright"
        footer_copyright = soup.select(footer_copyright_css_selector)
        if len(footer_copyright) == 0:
            raise CategoryHomePageContentError(
                "No footer copyright found with selector "
                f"'{footer_copyright_css_selector}'"
            )
        main_footer_copyright = None
        for fc in footer_copyright:
            if len(fc.text.strip()) > 0:
                if main_footer_copyright is None:
                    main_footer_copyright = fc
                else:
                    raise CategoryHomePageContentError(
                        "Too many footer copyright with selector "
                        f"'{footer_copyright_css_selector}'"
                    )
        if main_footer_copyright is None:
            raise CategoryHomePageContentError(
                "Empty footer copyright found with selector "
                f"'{footer_copyright_css_selector}'"
            )

        return f"{main_footer_copyright}"

    def get_online_metadata(self):
        """metadata from online website, looking at homepage source code"""
        logger.info("Fetching website metadata")

        soup, _ = get_soup("/")

        return {
            "title": soup.find("title").string,
            "description": soup.find("meta", attrs={"name": "description"}).attrs.get(
                "content"
            ),
            "footer_stats": self._extract_footer_stats_from_page(soup),
            "footer_copyright": self._extract_footer_copyright_from_page(soup),
        }



import traceback
from abc import ABC, abstractmethod
from queue import Queue

from .exceptions import FinalScrapingFailure
from .shared import Global, logger


class ScraperGeneric(ABC):
    def __init__(self):
        self.expected_items_keys = dict()
        self.unexpected_items_keys = dict()
        self.items_queue = Queue()
        self.missing_items_keys = set()
        self.error_items_keys = set()

    @abstractmethod
    def setup(self):
        pass

    @abstractmethod
    def get_items_name(self):
        pass

    @abstractmethod
    def build_expected_items(self):
        pass

    @abstractmethod
    def get_one_item_content(self, item_key, item_data):
        pass

    @abstractmethod
    def process_one_item(self, item_key, item_data, item_content):
        pass

    def add_item_to_scrape(self, item_key, item_data, is_expected):
        if (
            item_key in self.expected_items_keys
            or item_key in self.unexpected_items_keys
        ):
            return
        if is_expected:
            logger.debug(f"Adding {self.get_items_name()} {item_key} to scraping queue")
            self.expected_items_keys[item_key] = item_data
        else:
            logger.warning(
                f"Adding unexpected {self.get_items_name()} {item_key} "
                "to scraping queue"
            )
            self.unexpected_items_keys[item_key] = item_data
        self.items_queue.put(
            {
                "key": item_key,
                "data": item_data,
            }
        )

    def scrape_one_item(self, item_key, item_data):

        item_content = self.get_one_item_content(item_key, item_data)

        if item_content is None:
            logger.warning(f"Missing {self.get_items_name()} {item_key}")
            self.missing_items_keys.add(item_key)
            return

        logger.debug(f"Processing {self.get_items_name()} {item_key}")

        self.process_one_item(item_key, item_data, item_content)

    def scrape_items(self):

        logger.info(
            f"Scraping {self.get_items_name()} items ({self.items_queue.qsize()} "
            "expected)"
        )

        num_items = 1
        while not self.items_queue.empty():
            if Global.conf.scrape_only_first_items and num_items > 5:
                break
            try:
                item = self.items_queue.get(block=False)
                item_key = item["key"]
                item_data = item["data"]
                logger.info(
                    f"Scraping {self.get_items_name()} {item_key} ({num_items}/"
                    f"{len(self.expected_items_keys)+len(self.unexpected_items_keys)})"
                )
                self.scrape_one_item(item_key, item_data)
            except Exception as ex:
                self.error_items_keys.add(item_key)
                logger.warning(
                    f"Error while processing {self.get_items_name()} {item_key}: {ex}"
                )
                traceback.print_exc()
            finally:
                if (
                    len(self.missing_items_keys)
                    * 100
                    / (len(self.expected_items_keys) + len(self.unexpected_items_keys))
                    > Global.conf.max_missing_items_percent
                ):
                    raise FinalScrapingFailure(
                        f"Too many {self.get_items_name()}s found missing: "
                        f"{len(self.missing_items_keys)}"
                    )
                if (
                    len(self.error_items_keys)
                    * 100
                    / (len(self.expected_items_keys) + len(self.unexpected_items_keys))
                    > Global.conf.max_error_items_percent
                ):
                    raise FinalScrapingFailure(
                        f"Too many {self.get_items_name()}s failed to be processed: "
                        f"{len(self.error_items_keys)}"
                    )
                num_items += 1



# -*- coding: utf-8 -*-

import pathlib
import shutil
from datetime import datetime

from zimscraperlib.image.transformation import resize_image

from .constants import ROOT_DIR, Conf
from .scraper_category import ScraperCategory
from .scraper_guide import ScraperGuide
from .scraper_homepage import ScraperHomepage
from .scraper_info import ScraperInfo
from .shared import Global, GlobalMixin, logger
from .utils import setup_s3_and_check_credentials


class ifixit2zim(GlobalMixin):
    def __init__(self, **kwargs):

        Global.conf = Conf(**kwargs)
        for option in Global.conf.required:
            if getattr(Global.conf, option) is None:
                raise ValueError(f"Missing parameter `{option}`")

        self.scraper_homepage = ScraperHomepage()
        self.scraper_guide = ScraperGuide()
        self.scraper_category = ScraperCategory()
        self.scraper_info = ScraperInfo()
        self.scrapers = [
            self.scraper_homepage,
            self.scraper_category,
            self.scraper_guide,
            self.scraper_info,
        ]

    @property
    def build_dir(self):
        return self.conf.build_dir

    def cleanup(self):
        """Remove temp files and release resources before exiting"""
        if not self.conf.keep_build_dir:
            logger.debug(f"Removing {self.build_dir}")
            shutil.rmtree(self.build_dir, ignore_errors=True)

    def sanitize_inputs(self):
        """input & metadata sanitation"""
        logger.debug("Checking user-provided metadata")

        if not self.conf.name:
            is_selection = (
                self.conf.categories
                or self.conf.guides
                or self.conf.infos
                or self.conf.no_category
                or self.conf.no_guide
                or self.conf.no_info
            )
            self.conf.name = "ifixit_{lang}_{selection}".format(
                lang=self.conf.language["iso-639-1"],
                selection="selection" if is_selection else "all",
            )

        period = datetime.now().strftime("%Y-%m")
        if self.conf.fname:
            # make sure we were given a filename and not a path
            self.conf.fname = pathlib.Path(self.conf.fname.format(period=period))
            if pathlib.Path(self.conf.fname.name) != self.conf.fname:
                raise ValueError(f"filename is not a filename: {self.conf.fname}")
        else:
            self.conf.fname = f"{self.conf.name}_{period}.zim"

        if not self.conf.title:
            self.conf.title = self.metadata["title"]
        self.conf.title = self.conf.title.strip()

        if not self.conf.description:
            self.conf.description = self.metadata["description"]
        self.conf.description = self.conf.description.strip()

        if not self.conf.author:
            self.conf.author = "iFixit"
        self.conf.author = self.conf.author.strip()

        if not self.conf.publisher:
            self.conf.publisher = "openZIM"
        self.conf.publisher = self.conf.publisher.strip()

        self.conf.tags = list(
            set(
                self.conf.tag
                + ["_category:iFixit", "iFixit", "_videos:yes", "_pictures:yes"]
            )
        )

        logger.debug(
            "Configuration after sanitization:\n"
            f"name: {self.conf.name}\n"
            f"fname: {self.conf.fname}\n"
            f"name: {self.conf.author}\n"
            f"fname: {self.conf.publisher}"
        )

    def add_assets(self):
        """download and add site-wide assets, identified in metadata step"""
        logger.info("Adding assets")

        # recursively add our assets, at a path identical to position in repo
        assets_root = pathlib.Path(ROOT_DIR.joinpath("assets"))
        for fpath in assets_root.glob("**/*"):
            if not fpath.is_file():
                continue
            path = str(fpath.relative_to(ROOT_DIR))

            logger.debug(f"> {path}")
            with self.lock:
                self.creator.add_item_for(path=path, fpath=fpath)

    def add_illustrations(self):
        logger.info("Adding illustrations")

        src_illus_fpath = pathlib.Path(ROOT_DIR.joinpath("assets", "illustration.png"))
        tmp_illus_fpath = pathlib.Path(self.build_dir, "illustration.png")

        shutil.copy(src_illus_fpath, tmp_illus_fpath)

        # resize to appropriate size (ZIM uses 48x48 so we double for retina)
        for size in (96, 48):
            resize_image(tmp_illus_fpath, width=size, height=size, method="thumbnail")
            with open(tmp_illus_fpath, "rb") as fh:
                with self.lock:
                    self.creator.add_illustration(size, fh.read())

    def run(self):
        s3_storage = (
            setup_s3_and_check_credentials(self.conf.s3_url_with_credentials)
            if self.conf.s3_url_with_credentials
            else None
        )
        s3_msg = (
            f"\n"
            f"  using cache: {s3_storage.url.netloc} "
            f"with bucket: {s3_storage.bucket_name}"
            if s3_storage
            else ""
        )
        del s3_storage

        logger.info(
            f"Starting scraper with:\n"
            f"  language: {self.conf.language['english']}"
            f" ({self.conf.domain})\n"
            f"  output_dir: {self.conf.output_dir}\n"
            f"  build_dir: {self.build_dir}\n"
            f"{s3_msg}"
        )

        Global.metadata = self.scraper_homepage.get_online_metadata()
        logger.debug(
            f"Additional metadata scrapped online:\n"
            f"title: {self.metadata['title']}\n"
            f"description: {self.metadata['description']}\n"
            f"footer_stats: {self.metadata['footer_stats']}\n"
            f"footer_copyright: {self.metadata['footer_copyright']}\n"
        )
        self.sanitize_inputs()

        logger.debug("Starting Zim creation")
        Global.setup()
        Global.env.filters[
            "get_category_link"
        ] = self.scraper_category.get_category_link
        Global.env.filters["get_guide_link"] = self.scraper_guide.get_guide_link
        Global.env.filters["get_info_link"] = self.scraper_info.get_info_link
        Global.get_category_link = self.scraper_category.get_category_link
        Global.get_guide_link = self.scraper_guide.get_guide_link
        Global.get_info_link = self.scraper_info.get_info_link
        for scraper in self.scrapers:
            scraper.setup()
        self.creator.start()

        try:

            self.add_assets()
            self.add_illustrations()

            for scraper in self.scrapers:
                scraper.build_expected_items()

            while True:
                for scraper in self.scrapers:
                    scraper.scrape_items()
                needs_rerun = False
                if not Global.conf.scrape_only_first_items:
                    for scraper in self.scrapers:
                        if not scraper.items_queue.empty():
                            needs_rerun = True
                if not needs_rerun:
                    break

            logger.info("Awaiting images")
            Global.img_executor.shutdown()

            stats = "Stats: "
            for scraper in self.scrapers:
                stats += (
                    f"{len(scraper.expected_items_keys)} {scraper.get_items_name()}, "
                )
            for scraper in self.scrapers:
                stats += (
                    f"{len(scraper.missing_items_keys)} missing"
                    f" {scraper.get_items_name()}, "
                )
            for scraper in self.scrapers:
                stats += (
                    f"{len(scraper.error_items_keys)} {scraper.get_items_name()}"
                    " in error, "
                )
            stats += f"{self.imager.nb_requested} images"

            logger.info(stats)

        except Exception as exc:
            # request Creator not to create a ZIM file on finish
            self.creator.can_finish = False
            if isinstance(exc, KeyboardInterrupt):
                logger.error("KeyboardInterrupt, exiting.")
            else:
                logger.error(f"Interrupting process due to error: {exc}")
                logger.exception(exc)
            self.imager.abort()
            Global.img_executor.shutdown(wait=False)
            return 1
        else:
            logger.info("Finishing ZIM file")
            # we need to release libzim's resources.
            # currently does nothing but crash if can_finish=False
            # but that's awaiting impl. at libkiwix level
            with self.lock:
                self.creator.finish()
            logger.info(
                f"Finished Zim {self.creator.filename.name} "
                f"in {self.creator.filename.parent}"
            )
        finally:
            self.cleanup()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4 nu

import io
import locale
import re
import threading
import urllib.parse
import zlib
from contextlib import contextmanager
from typing import Union

import backoff
import bs4
import requests
from kiwixstorage import KiwixStorage
from pif import get_public_ip
from zimscraperlib.download import _get_retry_adapter, stream_file

from .constants import API_PREFIX
from .shared import Global, logger

LOCALE_LOCK = threading.Lock()


def to_path(url: str) -> str:
    """Path-part of an URL, without leading slash"""
    return re.sub(r"^/", "", urllib.parse.urlparse(url).path)


def get_url(path: str, **params) -> str:
    """url-encoded in-source website url for a path"""
    params_str = f"?{urllib.parse.urlencode(params)}" if params else ""
    return f"{Global.conf.main_url.geturl()}{urllib.parse.quote(path)}{params_str}"


def get_url_raw(path: str):
    """in-source website url for a path, untainted"""
    return f"{Global.conf.main_url.geturl()}{path}"


def to_url(value: str) -> str:
    """resolved potentially relative url from in-source link"""
    return value if value.startswith("http") else get_url_raw(value)


def to_rel(url: str) -> Union[None, str]:
    """path from URL if on our main domain, else None"""
    uri = urllib.parse.urlparse(url)
    if uri.netloc != Global.conf.domain:
        return None
    return uri.path


def no_leading_slash(text: str) -> str:
    """text with leading slash removed if present"""
    return re.sub(r"^/", "", text)


def no_trailing_slash(text: str) -> str:
    """text with trailing slash removed if present"""
    return re.sub(r"/$", "", text)


def only_path_of(url: str):
    """normalized path part of an url"""
    return normalize_ident(urllib.parse.urlparse(url).path)


def fetch(path: str, **params) -> str:
    """(source text, actual_paths) of a path from source website

    actual_paths is amn ordered list of paths that were traversed to get to content.
    Without redirection, it should be a single path, equal to request
    Final, target path is always last"""
    session = requests.Session()
    session.mount("http", _get_retry_adapter(10))  # tied to http and https
    resp = session.get(get_url(path, **params), params=params)
    resp.raise_for_status()

    # we have params meaning we requested a page (?pg=xxx)
    # assumption: this must be a category page (so on same domain)
    # we thus need to use redirection target (which lost param) with params
    if params and resp.history:
        return fetch(only_path_of(resp.url), **params)
    return resp.text, [
        no_leading_slash(only_path_of(r.url)) for r in resp.history + [resp]
    ]


def get_soup_of(text: str, unwrap: bool = False):
    """an lxml soup of an HTML string"""
    soup = bs4.BeautifulSoup(text, "lxml")
    if unwrap:
        for elem in ("body", "html"):
            getattr(soup, elem).unwrap()
    return soup


def get_soup(path: str, **params) -> bs4.BeautifulSoup:
    """an lxml soup of a path on source website"""
    content, paths = fetch(path, **params)
    return get_soup_of(content), paths


def get_digest(url: str) -> str:
    """simple digest of an url for mapping purpose"""
    return str(zlib.adler32(url.encode("UTF-8")))


def normalize_ident(ident: str) -> str:
    """URL-decoded category identifier"""
    return urllib.parse.unquote(ident)


def get_version_ident_for(url: str) -> str:
    """~version~ of the URL data to use for comparisons. Built from headers"""
    try:
        resp = requests.head(url)
        headers = resp.headers
    except Exception as exc:
        logger.warning(f"Unable to HEAD {url}")
        logger.exception(exc)
        try:
            _, headers = stream_file(
                url=url,
                byte_stream=io.BytesIO(),
                block_size=1,
                only_first_block=True,
            )
        except Exception as exc2:
            logger.warning(f"Unable to query image at {url}")
            logger.exception(exc2)
            return

    for header in ("ETag", "Last-Modified", "Content-Length"):
        if headers.get(header):
            return headers.get(header)

    return "-1"


def setup_s3_and_check_credentials(s3_url_with_credentials):
    logger.info("testing S3 Optimization Cache credentials")
    s3_storage = KiwixStorage(s3_url_with_credentials)
    if not s3_storage.check_credentials(
        list_buckets=True, bucket=True, write=True, read=True, failsafe=True
    ):
        logger.error("S3 cache connection error testing permissions.")
        logger.error(f"  Server: {s3_storage.url.netloc}")
        logger.error(f"  Bucket: {s3_storage.bucket_name}")
        logger.error(f"  Key ID: {s3_storage.params.get('keyid')}")
        logger.error(f"  Public IP: {get_public_ip()}")
        raise ValueError("Unable to connect to Optimization Cache. Check its URL.")
    return s3_storage


@contextmanager
def setlocale(name):
    with LOCALE_LOCK:
        saved = locale.setlocale(locale.LC_ALL)
        try:
            yield locale.setlocale(locale.LC_ALL, name)
        finally:
            locale.setlocale(locale.LC_ALL, saved)


def backoff_hdlr(details):
    logger.warning(
        "Backing off {wait:0.1f} seconds after {tries} tries "
        "calling function {target} with args {args} and kwargs "
        "{kwargs}".format(**details)
    )


@backoff.on_exception(
    backoff.expo,
    requests.exceptions.RequestException,
    max_time=16,
    on_backoff=backoff_hdlr,
)
def get_api_content(path, **params):
    full_path = get_url(API_PREFIX + path, **params)
    logger.debug(f"Retrieving {full_path}")
    response = requests.get(full_path)
    json_data = response.json() if response and response.status_code == 200 else None
    return json_data
